class Methods {
    static var Plurality = {
        name: 'Plurality',
        description: 'candidate who is the favourite of the most voters wins',
        elect: plurality, explain: explain_plurality};
    static var Condorcet = {
        name: 'Condorcet',
        description: 'candidate who is preferred to all others wins',
        elect: nearest, explain: explain_condorcet};
    static var Hare = {
        name: 'Hare',
        description: 'candidate who survives instant-runoff elimination wins',
        elect: hare, explain: explain_hare};
    static var Borda = {
        name: 'Borda',
        description: 'candidate with highest total ranking score wins',
        elect: borda, explain: explain_borda};
    static var Approval = {
        name: 'Approval',
        description: 'candidate acceptable to the most voters wins',
        elect: nearest, explain: explain_approval};

    static function find_min(array) {
        var min = array[0], index = 0;
        for (var i = 1; i < array.length; i++) {
            if (array[i] < min) {
                index = i;
                min = array[i];
            }
        }
        return index;
    }

    static function find_max(array) {
        var max = array[0], index = 0;
        for (var i = 1; i < array.length; i++) {
            if (array[i] > max) {
                index = i;
                max = array[i];
            }
        }
        return index;
    }
    
    static function count_firsts(votes, candidates) {
        var totals = [0, 0, 0, 0, 0];
        totals.length = candidates.length;
        for (var i = 0; i < votes.length; i++) {
            var count = votes[i].count;
            var ranks = votes[i].ranks;
            for (var j = 0; j < ranks.length; j++) {
                if (ranks[j] === 0) {
                    totals[j] += count;
                    break;
                }
            }
        }
        return totals;
    }
    
    static function plurality(votes, candidates, centroid, data) {
        var totals = count_firsts(votes, candidates);
        if (data) data.totals = totals;
        return candidates[find_max(totals)];
    }
    
    static function borda(votes, candidates, centroid, data) {
        var totals = [0, 0, 0, 0, 0];
        totals.length = candidates.length;
        var max_rank = candidates.length - 1;
        for (var i = 0; i < votes.length; i++) {
            var count = votes[i].count;
            var ranks = votes[i].ranks;
            for (var j = 0; j < ranks.length; j++) {
                totals[j] += (max_rank - ranks[j])*count;
            }
        }

        if (data) data.totals = totals;
        return candidates[find_max(totals)];
    }
    
    static function hare(votes, candidates, centroid, data) {
        var eliminated = [];
        if (data) data.events = [];
        while (true) {
            var totals = count_firsts(votes, candidates);
            var total = 0;
            for (var i = 0; i < candidates.length; i++) {
                total += totals[i];
            }
            for (var i = 0; i < candidates.length; i++) {
                if (totals[i]*2 > total) {
                    if (data) data.events.push(
                        candidates[i].name + ' wins with ' +
                        percent(totals[i]));
                    return candidates[i];
                }
            }
            for (var i = 0; i < eliminated.length; i++) {
                totals[eliminated[i]] = 1000;
            }
            var loser = find_min(totals);
            if (data) data.events.push(
                'eliminate ' + candidates[loser].name +
                ' with ' + percent(totals[loser]));
            for (var i = 0; i < votes.length; i++) {
                var ranks = votes[i].ranks;
                for (var j = 0; j < ranks.length; j++) {
                    if (ranks[j] > ranks[loser]) {
                        ranks[j]--;
                    }
                }
                ranks[loser] = 1000;
            }
            eliminated.push(loser);
        }
    }
    
    static function nearest(votes, candidates, centroid, data) {
        var distances = [];
        for (var i = 0; i < candidates.length; i++) {
            distances.push(Math.abs(candidates[i].position - centroid));
        }
        var winner_index = find_min(distances);
        data.winner_index = winner_index;
        return candidates[winner_index];
    }

    static function explain_totals(totals, candidates) {
        var display = [];
        for (var c = 0; c < candidates.length; c++) {
            display.push(candidates[c].name + ':\t' + percent(totals[c]));
        }
        return display;
    }

    static function score(n) {
        var value = int(n*1000);
        var frac = '.' + (value % 10);
        var whole = int(value/10);
        if (whole < 10) {
            whole = '    ' + whole;
        } else if (whole < 100) {
            whole = '  ' + whole;
        }
        return whole + frac;
    }

    static function percent(n) {
        var value = int(n*1000);
        var frac = '.' + (value % 10) + '%';
        var whole = int(value/10);
        if (whole < 10) {
            whole = '  ' + whole;
        }
        return whole + frac;
    }

    static function explain_plurality(votes, candidates, centroid, data) {
        return explain_totals(data.totals, candidates);
    }

    static function explain_approval(
        votes, candidates, centroid, data, distribution) {
        var totals = [];
        var cdf = distribution.cumulative;
        for (var c = 0; c < candidates.length; c++) {
            var last = cdf(candidates[c].position - 1 - centroid);
            var next = cdf(candidates[c].position + 1 - centroid);
            totals[c] = next - last;
        }
        return explain_totals(totals, candidates);
    }

    static function explain_condorcet(votes, candidates, centroid, data) {
        var display = [];
        var w = data.winner_index;
        var winner = candidates[w];
        for (var c = 0; c < candidates.length; c++) {
            if (c != w) {
                var line = winner.name + ' > ' + candidates[c].name + ':\t';
                var total = 0;
                for (var v = 0; v < votes.length; v++) {
                    if (votes[v].ranks[w] < votes[v].ranks[c]) {
                        total += votes[v].count;
                    }
                }
                line += percent(total);
                display.push(line);
            }
        }
        return display;
    }

    static function explain_hare(votes, candidates, centroid, data) {
        return data.events;
    }

    static function explain_borda(votes, candidates, centroid, data) {
        var display = [];
        for (var c = 0; c < candidates.length; c++) {
            display.push(candidates[c].name + ':\t' + score(data.totals[c]));
        }
        return display;
    }
}
